
/*
 * Autosscep -- Simple SCEP client implementation
 * Copyright (c) Alberto Forino 2004. All rights reserved.
 * See the file COPYRIGHT for licensing information.
 */


/* Main routine */

#include <time.h>
#include <math.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/asn1.h>

#include "autosscep.h"

#define	VERSION	"0.9.28b  20 September 2004"

/* Main routine */

int
main(int argc, char **argv)
{
	int             cert_update_flag = 0;
	struct ca_data *temp;
	int             c,  i = 0, k = 0;
	char           *p;//*aux;
	char            http_string[16384];
	struct http_reply reply;
	struct scep     scep_t;
	FILE           *fp = NULL;

	/* initialize the ca struct vector */
	*CAs = NULL;
	/* initialize the flag-counters */
	cert_flag = 0;
	ca_flag = 0;

	/* Set program name */
	pname = argv[0];

	/* Define signal trap */
	(void) signal(SIGALRM, catchalarm);

	/* Set timeout */
	timeout = TIMEOUT;

        /* Check of argument */
        if( argc < 2 || !strcmp(argv[1],"-h")|| !strcmp(argv[1],"--help" )){
             printf ("AutoSscep, Automatic Simple certificate enroller\n"
                  "\n\tsyntax:  autosscep configuration_file \n"
                  "\nRead the README file for other help,\nor the HOWTOCONFIGURE manual to fill the configuration file."
                  "\nIf you haven't these file download them from http://autosscep.spe.net\n");
           exit(0);
        }else if (!strcmp(argv[1],"-version")){     /* option for display program version*/
           printf ("AutoSScep %s\n", VERSION);
           exit(0);
        }

        /* Set the reading of config file */
        f_char = argv[1];

	/* Read in the configuration file: */
	if (!(fp = fopen(f_char, "r"))) {
		fprintf(stderr, "%s: cannot open the configuration file: %s\n",
			pname, f_char);
		exit(-1);
	} else {
		/* read */
		init_config(fp);
		/* check */
		check_config();
		(void) fclose(fp);
	}

	if (v_flag)
		fprintf(stdout, "%s: starting autoscep, version %s\n", pname, VERSION);

	/* Comincia il loop sui certificati */


enroll:
	for (; i < cert_flag; i++) {	/* per tutti i certificati specificati nel file di config */

		while (certs[i]->err_flag < 0)
			i++;	/* salta i certificati che hanno riscontrato errori gravi */
		fprintf(stdout, "%s: Checking certificate -- > %s\n", pname, certs[i]->cert_file);

		if (!check_date(certs[i]->cert, days)) {
			fprintf(stdout, "%s: Certificate '%s' doesn't need to be enrolled yet\n", pname, certs[i]->cert_file);
		} else {
			//bisogna rinnovare
				if (v_flag) {
				fprintf(stdout, "%s: Certificate %s is going to expire (or is missing)\n", pname, certs[i]->cert_file);
				fprintf(stdout, "%s: Looking for the CA data \n", pname);
			}
			temp = get_ca_struct(certs[i]->cert);

			if (temp == NULL)	/* mancando il certificato cerca i dati della ca specificati nel file config  */
				for (k = 0; k < ca_flag; k++) {
					if (!strcmp(certs[i]->CADN, CAs[k]->DN)) {
						temp = CAs[k];
						k = ca_flag;
					}
				}
			/*
			 * temp punta ai dati della CA firmataria del certificato
			 */

			if (temp == NULL) {
				fprintf(stderr, "%s: ERROR: no CA specified for the certificate,: %s\n\n", pname, certs[i]->cert_file);
				certs[i]->err_flag = ERR_CERT_MISSING_CA;
			} else {
				/* una volta trovata la ca e i suoi dati si pu iniziare l'enrollment */
				if (temp->err_flag < 0) {
					i++;
					goto enroll;
				}
				if (v_flag)
					fprintf(stdout, "%s: CA founded!! \n%s: Starting certificate enrollment for -- > %s\n", pname, pname, certs[i]->cert_file);
				/* controlla che il certificato non sia gi stato richiesto con una req pendente */
				if ((ctrl_pending(certs[i])) != NULL) {
					printf("%s: Pending request \n", pname);
					/* imposta la transazione per la  pending e legge la chiave (che era stata generata nuova) */
                                       certs[i]->transactionID=ctrl_pending(certs[i]);
                                       snprintf(temp_string,temp_string_len,"%s%s.new",key_path,certs[i]->key_file);
                                       certs[i]->new_key = get_key(temp_string);
                                       	if (certs[i]->new_key == NULL) {
					/* controlla se la lettura della chiave ha dato errore */
						fprintf(stderr, "%s: Missing private key file '%s'\n", pname, temp_string);
						certs[i]->err_flag = ERR_CERT_KEY_MISSING;
						i++;
          	                                goto enroll;
					}

					/* legge la req */
					snprintf(temp_string,temp_string_len,"%s%s.csr",cert_path,certs[i]->cert_file);
   				        certs[i]->req = get_req(temp_string);

					if (certs[i]->req == NULL) {
					/* controlla se la lettura della cert req ha dato errore */
						certs[i]->err_flag = ERR_CERT_REQ_MISSING;
						i++;
						goto enroll;
					}
					/* imposta la transazione */
					new_transaction(&scep_t, certs[i]);
     					init_scep();
					scep_t.transaction_id = ctrl_pending(certs[i]);
					scep_t.request_type = SCEP_REQUEST_GETCERTINIT;
					scep_t.request_type_str = SCEP_REQUEST_GETCERTINIT_STR;

				} else {
                                        /* imposta la transazione per una richiesta di cert nuova */
					printf("%s: New request \n", pname);

					/* genera una nuova chiave privata e la mette nella struttura del cert*/
					if (gen_new_key(certs[i]) < 0) {
						fprintf(stderr, "%s: ERROR ON WRITING PRIVATE KEY, copy it from here\n", pname);
						PEM_write_PrivateKey(stderr, certs[i]->new_key, NULL, NULL, 0, NULL, NULL);
					}
					/* genera la nuova richiesta di certificato */
					gen_new_req(certs[i]);
					if (certs[i]->err_flag < 0) {
						i++;
						goto enroll;
					}
					/* inizializza lo scep */
					init_scep();

					/* crea una nuova transazione scep */
					new_transaction(&scep_t, certs[i]);
                                        /*copia il transaction id nella struttura del cert */
                                        certs[i]->transactionID=strdup(scep_t.transaction_id);
					/* imposta il tipo di messaggio scep  --> PKCSReq (19)	 */
					scep_t.request_type = SCEP_REQUEST_PKCSREQ;
				}

				/* firma la transazione con un certificato autofirmato */
				if (v_flag)
					fprintf(stdout, "%s: generating selfsigned certificate\n", pname);

				scep_t.selfsigned = new_selfsigned(certs[i]);
				scep_t.pkey = certs[i]->new_key;

				/* imposta il subject della PKCS7 com quello della req cert */
				if (!(scep_t.ias_getcertinit->subject = X509_REQ_get_subject_name(certs[i]->req))) {
					fprintf(stderr, "%s: error getting subject for GetCertInitial\n", pname);
					ERR_print_errors_fp(stderr);
					certs[i]->err_flag = ERR_CERT_BAD_REQ;
				}
				if (certs[i]->err_flag < 0) {
				/* se si sono verificati errori inizia l'enroll di un altro certificato */
					i++;
					goto enroll;
				}
				printf("%s: sending certificate request \n%s: transaction_ID = %s\n", pname, pname, scep_t.transaction_id);

				/* create payload */
				pkcs7_wrap(&scep_t, certs[i], temp);

				/* URL-encode */
				p = url_encode(scep_t.request_payload, scep_t.request_len);

				/* Forge the HTTP message */
                                if (temp->VHost_flag == 1)	/* se il server della ca ha dei virtual host */
					snprintf(http_string, sizeof(http_string),
						 "GET %s%s?"
						 "operation=PKIOperation&"
						 "message="
					  "%s HTTP/1.0\r\nHost: %s\r\n\r\n",
					         p_flag ? "" : "/",temp->host_dir, p, temp->host);
				else
					snprintf(http_string, sizeof(http_string),
						 "GET %s%s?operation="
						 "PKIOperation&message="
						 "%s HTTP/1.0\r\n\r\n",
						 p_flag ? "" : "/", temp->host_dir, p);

				if (d_flag)
					fprintf(stdout, "%s: scep msg: %s", pname, http_string);

				/* send http (reply in http reply struct 'reply' ) */
				reply.payload = NULL;
				if ((c = send_msg(&reply, http_string, temp->host, atoi(temp->port), operation_flag)) == 1) {
					fprintf(stderr, "%s: error while sending " "message\n", pname);
					certs[i]->err_flag = ERR_CERT_HTTP;
					i++;
					goto enroll;
				}
				/* Verisign Onsite returns strange reply for pending */
				if ((reply.status == 200) && (reply.payload == NULL)) {
					scep_t.pki_status = SCEP_PKISTATUS_PENDING;
					if (ctrl_pending(certs[i])==NULL){
                                           write_pending(certs[i]);
                                           system(certs[i]->req_command);
                                        }
                                        certs[i]->renewed_flag = scep_t.pki_status;
					break;
				}
				printf("%s: valid response from server\n", pname);

				/* Check payload */
				scep_t.reply_len = reply.bytes;
				scep_t.reply_payload = reply.payload;
				pkcs7_unwrap(&scep_t, certs[i], temp);
				pkistatus = scep_t.pki_status;
				certs[i]->renewed_flag = scep_t.pki_status;
				switch (scep_t.pki_status) {

				case SCEP_PKISTATUS_SUCCESS:
					write_local_cert(&scep_t, certs[i]);
					/* cancellazione file pending */
					snprintf(temp_string,temp_string_len,"%s%s.pending",cert_path,certs[i]->cert_file);
          remove(temp_string);

					break;

				case SCEP_PKISTATUS_PENDING:
					/* Check time limits */
					certs[i]->transactionID = strdup(scep_t.transaction_id);
					/* scrive il transaction id nel file pending */
					if (ctrl_pending(certs[i])==NULL){
                                           write_pending(certs[i]);
                                           system(certs[i]->req_command);
                                        }
                                        break;

				case SCEP_PKISTATUS_FAILURE:
					printf("%s: Failure : %d\n", pname, scep_t.fail_info);
					break;

				default:
					fprintf(stderr, "%s: unknown " "pkiStatus\n", pname);
					certs[i]->err_flag = ERR_CERT_BAD_REPLY;
					break;

				}	/* switch */
			}	/* else ca non trovata */
		}		/* else bisogna rinnovare */
	}			/* for su certificati              */

	/* terminati tutti i certificati */

	/* controlla quali certificati sono stati rinnovati, li rinomina e visualizza un riassunto sulle operazioni fatte*/
	printf("%s: --- SUMMARY ---\n",pname);
	for (i = 0; i < cert_flag; i++) {
		printf("%s: Certificate : %s (status %d)\n", pname, certs[i]->cert_file,certs[i]->renewed_flag);
		if (certs[i]->renewed_flag == 0) {
			printf("\tCertificate updated\n");
			cert_update_flag++;
			update(certs[i]);

		} else if (certs[i]->renewed_flag == 3)
			printf("\tCertificate request is pending on ca/ra\n");
		else if (certs[i]->renewed_flag == 2)
			printf("\tSCEP transaction failed\n");
		if (certs[i]->err_flag < 0)
			printf("\terror : %d\n", certs[i]->err_flag);

	}

	printf("%s: Total updated certificates: %d\n", pname, cert_update_flag);

	/* Se  stato rinnovato qualche certificato lancia l'extra command  */
	if (cert_update_flag > 0 && e_command != NULL) {
		printf("%s: Executing '%s'\n", pname, e_command);
		system(e_command);
	}
	return 0;


}
void
catchalarm(int signo)
{
	fprintf(stderr, "%s: connection timed out\n", pname);
	exit(SCEP_PKISTATUS_TIMEOUT);
}
