
/*
 * Autosscep -- Simple SCEP client implementation
 * Copyright (c) Alberto Forino 2004. All rights reserved.
 * See the file COPYRIGHT for licensing information.
 */



/* certificates manipulation routines */

#include "autosscep.h"
#include <errno.h>


/* Open the inner, decrypted PKCS7 and try to write cert.  */
void
write_local_cert(struct scep * s, struct cert_data * ce)
{
	PKCS7          *p7;
	STACK_OF(X509) * certs;
	X509           *cert = NULL;
	char            CN_req[1024], CN_cert[1024];
	int             i;
	X509           *localcert;

	localcert = NULL;

	/* Get certs */
	p7 = s->reply_p7;
	certs = p7->d.sign->cert;

	/* Find cert */
	for (i = 0; i < sk_X509_num(certs); i++) {
		char            buffer[1024];

		cert = sk_X509_value(certs, i);
		if (v_flag) {
			printf("%s: found certificate with\n" "  subject: %s\n", pname,
			       X509_NAME_oneline(X509_get_subject_name(cert), buffer, sizeof(buffer)));
			printf("  issuer: %s\n", X509_NAME_oneline(X509_get_issuer_name(cert), buffer, sizeof(buffer)));
		}
		/* The subject has to match that of our request control of CN */

		//extract CN from req and put it in CN_req
			X509_NAME_get_text_by_NID(X509_REQ_get_subject_name(ce->req), NID_commonName, CN_req, 256);
		//extract CN from cert and put it in CN_cert
			X509_NAME_get_text_by_NID(X509_get_subject_name(cert), NID_commonName, CN_cert, 256);

		if (!strcmp(CN_req, CN_cert)) {
			if (v_flag)
				printf("%s: CN match !!\n", pname);
			localcert = cert;
		}

	}
	if (localcert == NULL) {
		fprintf(stderr, "%s: cannot find requested certificate\n", pname);

	}
	/* Write PEM-formatted file: */

	/* set up nome */
  snprintf(temp_string,temp_string_len,"%s%s.new",cert_path,ce->cert_file);

	/* write cert */
	if (v_flag)
		printf("%s: writing cert \n", pname);
	write_cert(localcert, temp_string);

	if (d_flag)
		PEM_write_X509(stdout, localcert);
	printf("%s: certificate written as %s\n", pname, temp_string);

}


/*
 * Write on file [nomefile] the certificate
 */
int
write_cert(X509 * x, char *nomefile)
{
	FILE           *fp = NULL;
	/* open file pointer */
	if (!(fp = fopen(nomefile, "w"))) {
		fprintf(stderr, "%s: cannot open cert file for writing\n", pname);
		return -1;
	}
	/* write certificate */
	if (PEM_write_X509(fp, x) != 1) {
		fprintf(stderr, "%s: error while writing certificate " "file\n", pname);
		ERR_print_errors_fp(stderr);
		return -1;
	}
	fclose(fp);
	return 1;
}



/*
 * Return the certificate contained in the specified file (with the cert * path)
 */
X509           *
get_cert(char *nomefile)
{
	X509           *cert = NULL;
	FILE           *file_p;
	file_p = NULL;

	//apre il file in lettura
		if (!(file_p = fopen(nomefile, "r"))) {
		fprintf(stderr, "%s: cannot open local file: '%s'\n", pname, nomefile);
		return NULL;
	}
	if (v_flag)
		fprintf(stdout, "%s: File '%s' opened\n", pname, nomefile);

	//Legge il certificato dal file
		if (!(PEM_read_X509(file_p, &cert, NULL, NULL))) {
		fprintf(stderr,
			"%s: error while reading local cert (maybe it isn't a PEM format cert)\n", pname);
		ERR_print_errors_fp(stderr);
		fclose(file_p);
		return NULL;
	}
	fclose(file_p);

	return cert;
}


/* Legge chiave privata   da file */

EVP_PKEY       *
get_key(char *s)
{
	FILE           *fp;
	EVP_PKEY       *evp_k;
	if (!(fp = fopen(s, "r")))
		return NULL;
	evp_k = EVP_PKEY_new();
	if (!PEM_read_PrivateKey(fp, &evp_k, NULL, NULL)) {
		fprintf(stderr, "%s: error while reading private key file '%s'\n", pname, s);
		ERR_print_errors_fp(stderr);
		return NULL;
	}
	fclose(fp);
	return evp_k;
}

/*
 * Ritorna 1 se il certificato passato  usato per crittare i dati
 */

int
check_enc(X509 * cert)
{
	int             index;
	X509_EXTENSION *ext;
	ASN1_OCTET_STRING *key_usage;

	index = X509_get_ext_by_NID(cert, NID_key_usage, -1);
	ext = X509_get_ext(cert, index);
	X509V3_EXT_print_fp(stdout, ext, 0, 0);
	key_usage = (ASN1_BIT_STRING *) X509V3_EXT_d2i(ext);
	return ASN1_BIT_STRING_get_bit(key_usage, 2);
}

/* Ritorna 1 se il certificato passato  usato per firmare */
int
check_sig(X509 * cert)
{
	int             index;
	X509_EXTENSION *ext;
	ASN1_OCTET_STRING *key_usage;

	index = X509_get_ext_by_NID(cert, NID_key_usage, -1);
	ext = X509_get_ext(cert, index);
	key_usage = (ASN1_BIT_STRING *) X509V3_EXT_d2i(ext);
	return ASN1_BIT_STRING_get_bit(key_usage, 0);
}


/*
 * Open the inner, decrypted PKCS7 and try to write CA/RA certificates
 */
int
write_ca_ra(struct http_reply * s, struct ca_data * ca)
{
	BIO            *bio;
	PKCS7          *p7;
	STACK_OF(X509) * certs = NULL;
	X509           *cert = NULL;
	int             i;
	char            buffer[1024];
	char            name[1024];

	/* Create read-only memory bio */
	bio = BIO_new_mem_buf(s->payload, s->bytes);
	p7 = d2i_PKCS7_bio(bio, NULL);
	if (p7 == NULL) {
		fprintf(stderr, "%s: error reading PKCS#7 data\n", pname);
		ERR_print_errors_fp(stderr);
		ca->err_flag = ERR_CA_FAILED_CACERT_REQ;
		return -1;
	}

	/* Get certs */
	i = OBJ_obj2nid(p7->type);
	switch (i) {
	case NID_pkcs7_signed:
		certs = p7->d.sign->cert;
		break;
	default:
		printf("%s: wrong PKCS#7 type\n", pname);
		ca->err_flag = ERR_CA_FAILED_CACERT_REQ;
		return -1;
	}
	/* Check  */
	if (certs == NULL) {
		fprintf(stderr, "%s: cannot find certificates\n", pname);
		ca->err_flag = ERR_CA_FAILED_CACERT_REQ;
		return -1;
	}
	for (i = 0; i < sk_X509_num(certs); i++) {
		cert = sk_X509_value(certs, i);
		printf("\n%s: n.%d -> found certificate with\n  subject: %s\n", pname, i,
		       X509_NAME_oneline(X509_get_subject_name(cert), buffer, sizeof(buffer)));
		if (check_enc(cert) == 1) {
			//il certificato di crittaggio dei dati
				snprintf(name, 1024, "%s%s", ca_path, ca->enc_file);
			ca->enc_cert = cert;
			if (check_sig(cert) == 1) {
				//sia crittaggio che sig
					ca->cert = cert;
				//assegna lo stesso cert a ca->cert
			}
		} else if (check_sig(cert) == 1 && ca->cert == NULL) {
			ca->cert = cert;
			snprintf(name, 1024, "%s%s", ca_path, ca->cert_file);
		} else {
			snprintf(name, 1024, "%s%s-%d", ca_path, ca->cert_file, i);
		}

		write_cert(cert, name);

	}
	return 1;

}

/* legge la req da file */
X509_REQ       *
get_req(char *nomefile)
{
	X509_REQ       *x;
	FILE           *fp;
	if ((fp = fopen(nomefile, "r")) == NULL) {
		fprintf(stderr, "%s: cannot open %s file for reading\n", pname, nomefile);
		exit(1);
	}
 	/* lettura del certificato */
		x = PEM_read_X509_REQ(fp, NULL, NULL, NULL);
	return x;
}
