
/*
 * Autosscep -- Simple SCEP client implementation
 * Copyright (c) Alberto Forino
 * 2004. All rights reserved. See the file COPYRIGHT for licensing on.
 */ 

 /* Configuration file initialization */ 

#include "autosscep.h"
#include <string.h>

void           
check_general_option() 
{

	FILE * fp;
        /* controlla che le opzioni obbligatorie siano state inserite */
	if (!ca_path_flag) {
	   fprintf(stderr, "%s: You must declare the CA directory (CADir) before of any CA or certificate Data !!\n", pname);
	   exit(0);
	}
	if (!cert_path_flag) {
	   fprintf(stderr, "%s: You must declare the certificate directory (CertDir) before of any CA or certificate Data !!\n", pname);
	   exit(0);
	}
	if (!key_path_flag) {
  	   fprintf(stderr, "%s: You must declare the private keys' directory (KeyDir) before of any CA or certificate Data !!\n", pname);
	   exit(0);
	}
		
	/*
	 * controlla la possibilit di lettura scrittura nelle cartelle specificate
	 */

        //cert dir
	snprintf(temp_string,temp_string_len,"%s%s",cert_path,TEST_FILE_NAME);
		if (!(fp = fopen(temp_string, "w+"))) {
			fprintf(stderr, "%s: ERROR: no permission for writing on %s\n", pname, cert_path);
			exit(0);
      	        } else {
  			fclose(fp);
			remove(temp_string);
           	}
 	//key dir
		snprintf(temp_string,temp_string_len,"%s%s",key_path,TEST_FILE_NAME);
		if (!(fp = fopen(temp_string, "w+"))) {
		   fprintf(stderr, "%s: ERROR: no permission for writing on %s\n", pname, key_path);
		   exit(0);
      	        } else {
                  fclose(fp);
  		  remove(temp_string);
          	}

} 

/*
 * Legge il file di configurazioe e alloca i dati nelle strutture globali certs[] e CAs[]
 */
void           
init_config(FILE * conf) 
{
	char            buff[1024];
	char           *str2;
	int             i, lines;
	p_flag = 1;
	lines = 0;
	fprintf(stdout, "%s: Reading config file \n", pname);
	while (fgets(buff, 1024, conf))	/* ciclo su tutto il file */
	{
			lines++;
			 /* null-terminate: */
			buff[strlen(buff) - 1] = '\0';
			 /* skip leading white space: */
			for (i = 0; isspace(buff[i]); i++)
                            ;
           		 /* empty line? */
			if (!strlen(&buff[i]))
				continue;
			 /* comment? */
			if (!strncmp("#", &buff[i], 1))
				continue;
			if (!strncmp(&buff[i], "CertDir", 7))	/* campo cert dir */
                           {
      			   str2 = get_string(&buff[i]);
                           cert_path_flag = 1;
      			   cert_path = malloc(sizeof(char) * strlen(str2) +1);
			   strcpy(cert_path, str2);
      		        } else if (!strncmp(&buff[i], "KeyDir", 6))
                          {
                          str2 = get_string(&buff[i]);
                          key_path_flag = 1;
                          key_path = malloc(sizeof(char) * strlen(str2) +1);
                          strcpy(key_path, str2);
                        } else if (!strncmp(&buff[i], "CADir", 5))
                          {
                          str2 = get_string(&buff[i]);
                          ca_path_flag = 1;
                          ca_path = malloc(sizeof(char) * strlen(str2) +1);
                          strcpy(ca_path, str2);
                        } else if (!strncmp(&buff[i], "DaysBeforeExpire", 16))
                          {
                          str2 = get_string(&buff[i]);
                          days_flag = 1;
                          days = atoi(str2);
                        } else if (!strncmp(&buff[i], "Proxy", 5))
                          {
                          str2 = get_string(&buff[i]);
                          p_flag = 1;
                          if (!(p_char = strdup(str2)))
			     error_memory();
                        } else if (!strncmp(&buff[i], "KeyBits", 7))
                          {
                          str2 = get_string(&buff[i]);
                          kbits_flag = 1;
                          kbits = atoi(str2);
                        } else if (!strncmp(&buff[i], "[CA]", 4))
                          {
     			/*  controlla che tutte le opzioni principali
			 * arie siano gi state settate  */
			  check_general_option();

                        /* Inizializzazione della struttura della ca con i
			 * dati contenuti nel file config */
                         CAs[ca_flag] = malloc(sizeof(struct ca_data));
   			 set_ca_struct(conf, CAs[ca_flag]);
			 ca_flag++;
 		        } else if (!strncmp(&buff[i], "Verbose", 7))
                     	  {
                          str2 = get_string(&buff[i]);
                          if (!strncmp(str2, "yes", 3))
                          v_flag = 1;
                        } else if (!strncmp(&buff[i], "Debug", 5))
                          {
                          str2 = get_string(&buff[i]);
                          if (!strncmp(str2, "yes", 3))
                          d_flag = 1;
                        } else if (!strncmp(&buff[i], "SigAlgo", 7))
                          {
                          str2 = get_string(&buff[i]);
                          S_char = strdup(str2);
                          S_flag = 1;
                        } else if (!strncmp(&buff[i], "[Certificate]", 12))
                          {
                          check_general_option();
                          /* Crea la struttura contenente i dati del certificato */
		          certs[cert_flag] = malloc(sizeof(struct cert_data));
                          set_cert_struct(conf, certs[cert_flag]);
     			/* Legge il certificato dal file e lo mette nella struttura */
	                  cert_flag++;
                        } else if (!strncmp(&buff[i], "ECommand", 8))
                          {
                          e_command = get_string(&buff[i]);
                        } else
                          {
                          fprintf(stderr, "%s: UNKNOWN OPTION : '%s'\n", pname, &buff[i]);
                          }
        }
}

/*
 * ritorna un puntatore ad una copia della stringa limitata da apici contenuta
 * nella stringa passata
 */ 
char           *
get_string(char *str) 
{
	char           *temp;
		int             i, k;
		for (i = 0; str[i] != '"'; i++);	/* trova inizio */
		for (k = i + 1; str[k] != '"'; k++);	/* trova fine */
		temp = malloc(sizeof(char) * (1 + k - i - 1));
		strncpy(temp, str + i + 1, k - i - 1);
		temp[k - i - 1] = '\0';
		return temp;
}

void           
error_memory() 
{
	fprintf(stderr, "%s: memory allocation failure, errno: %d\n", pname, errno);
		exit(-1);
}

/*
 * Controlla i valori settati e mette di default quelli non specificati
 */ 
int            
check_config() 
{
	if (v_flag)
		fprintf(stdout, "%s: Checking config file values\n", pname);
		if (!cert_path_flag) {
			fprintf(stderr, "%s: no cert path specified in the config file\n", pname);
			exit(0);
	}
		if (!key_path_flag) {
			fprintf(stderr, "%s: no keys path specified in the config file\n", pname);
			exit(0);
	}
		if (!days_flag) {
			days = 7;	/* assegna  7 giorni di default */
	}
		if (!kbits_flag) {
			kbits = DEFAULT_BITS;	/* assegna i bit di default */
	}
		if (ca_flag < 1) {
			fprintf(stderr, "%s: you must specify at least one CA for your certificates in the config file\n", pname);
			exit(0);
	}
		if (cert_flag < 1) {
			fprintf(stderr, "%s: No certificate specified in the config file\n", pname);
			exit(0);
	}
		
		if (!S_flag) {
			if (v_flag)
				fprintf(stdout, "%s: Default signature algorithm : md5\n", pname);
			sig_alg = (EVP_MD *) EVP_md5();
	} else if (!strncmp(S_char, "md5", 3)) {
			if (v_flag)
				fprintf(stdout, "%s: Signature algorithm specified: %s\n", pname, S_char);
			sig_alg = (EVP_MD *) EVP_md5();
	} else if (!strncmp(S_char, "sha1", 4)) {
			if (v_flag)
				fprintf(stdout, "%s: Signature algorithm specified: %s\n", pname, S_char);
			sig_alg = (EVP_MD *) EVP_sha1();
	} else {
			fprintf(stderr, "%s: unsupported algorithm: '%s', md5 will be used\n", pname, S_char);
			sig_alg = (EVP_MD *) EVP_md5();
	}
		return 1;
}

/* Setta i dati di una ca partendo da una porzione del file di configurazione
* variabili globali usate : ca_path
 * 
 * parametri modificati : la struttura ca_data passata viene riempita in base ad
 *  campi del file di config
 */ 
int            
set_ca_struct(FILE * fp, struct ca_data * c) 
{
	 /* lettura del file config */
		char            buff[1024];
		int             i;
		 /* valori di default */
		c->DN = NULL;
		c->cert_file = NULL;
		c->enc_file = NULL;
		c->url_scep = NULL;
		c->VHost_flag = 0;
		c->err_flag = 0;
		c->ID = NULL;
		c->port="80";
		while (fgets(buff, 1024, fp) && strncmp(buff, "[/CA]", 5))
	              {
			 /* null-terminate: */
			buff[strlen(buff) - 1] = '\0';
			 /* skip leading white space: */
			for (i = 0; isspace(buff[i]); i++)
				;
			 /* empty line? */
			if (!strlen(&buff[i]))
				continue;
			 /* comment? */
			if (!strncmp("#", &buff[i], 1))
				continue;
			for (; isspace(buff[i]); i++);
			if (!strncmp(&buff[i], "DN", 2))
		{
				c->DN = get_string(&buff[i]);
		} else if (!strncmp(&buff[i], "URL", 3))
		{
				c->url_scep = get_string(&buff[i]);
		} else if (!strncmp(&buff[i], "CertFile", 8))
		{
				c->cert_file = get_string(&buff[i]);
		} else if (!strncmp(&buff[i], "EncCertFile", 11))
		{
				c->enc_file = get_string(&buff[i]);
		} else if (!strncmp(&buff[i], "EncAlgo", 7))
		{
				c->enc_alg_str = get_string(&buff[i]);
		} else if (!strncmp(&buff[i], "VHost", 5))
		{
				if (!strcmp(get_string(&buff[i]), "yes"))
					c->VHost_flag = 1;
		} else if (!strncmp(&buff[i], "CAIdentifier", 12))
		{
				c->ID = get_string(&buff[i]);
		} else
		{
				fprintf(stderr, "%s: UNKNOWN OPTION : '%s'\n", pname, &buff[i]);
		}
	}
		if (c->DN == NULL || c->cert_file == NULL || c->enc_file == NULL || c->url_scep == NULL)
	{
			c->err_flag = ERR_CA_BAD_CONFIG;
			fprintf(stderr,
				"%s: BAD CONFIGURATION OF CA DATA (something is missing), check the config file\n", pname);
			return -1;
	}
		 /* set up indirizzi */
		set_host_data(c);
		 /* setup algoritmo di cifratura */
		if (c->enc_alg_str == NULL) {
			c->enc_alg = (EVP_CIPHER *) EVP_des_cbc();	/* default = DES */
	} else if (!strncmp(c->enc_alg_str, "blowfish", 8))
	{
			c->enc_alg = (EVP_CIPHER *) EVP_bf_cbc();
	} else if (!strncmp(c->enc_alg_str, "des", 3))
	{
			c->enc_alg = (EVP_CIPHER *) EVP_des_cbc();
	} else if (!strncmp(c->enc_alg_str, "3des", 4))
	{
			c->enc_alg = (EVP_CIPHER *) EVP_des_ede3_cbc();
	} else if (c->enc_alg_str != NULL)
	{
			fprintf(stderr,
				"%s: CA %s error - unsupported algorithm: '%s', using DES\n", 
				pname, c->cert_file, c->enc_alg_str);
			c->enc_alg = (EVP_CIPHER *) EVP_des_cbc();
	}
		 /* caricamento dei certificati dal/dai file/s certificato (della CA)*/
		snprintf(temp_string,temp_string_len,"%s%s",ca_path,c->cert_file);
		c->cert = get_cert(temp_string);
		if (c->cert == NULL)
	{
			request_ca_cert(c);
			if (c->err_flag < 0) {
				fprintf(stderr, "%s: ERROR WHILE READING CA CERT FILE\n", pname);
				c->err_flag = ERR_CA_CERT_MISSING;
				return -1;
		}
			return 1;	/* richiesti i certificati ritorna
					 * l' init config */
	}

		 /* enc cert */
		snprintf(temp_string,temp_string_len,"%s%s",ca_path,c->enc_file);

		c->enc_cert = get_cert(temp_string);
  		if (c->enc_cert == NULL)
             	   {
     			fprintf(stderr, "%s: ERROR WHILE READING CA ENC CERT FILE\n", pname);
			c->err_flag = ERR_CA_ENC_CERT_MISSING;
			return -1;
	}
		return 1;
}

/* Setup the ca host data from the scep url */
int            
set_host_data(struct ca_data * c) 
{
	int             i,k;
		if (strncmp(c->url_scep, "http://", 7))
	{
			fprintf(stderr,
				"%s : Error : no valid url '%s' \n\t It should be an http address\n", 
				pname, c->url_scep);
	}
		 /* set up host */
		for (i = 0; c->url_scep[i + 7] != '/'; i++);
		c->host = malloc(sizeof(char) * ((i) + 1));
		strncpy(c->host, c->url_scep + 7, i);
		/*looking for port number*/
		for (k = 0; k< strlen(c->host); k++){
                if (c->host[k]==':')         {
                   c->port=malloc(strlen(c->host)-k+1);
                   strcpy(c->port,c->host+k+1);
                   c->host[k]='\0';
                   k=strlen(c->host)+1;
                   }
                }

		 /* set up host directory */
		c->host_dir = malloc(sizeof(char) * (strlen(c->url_scep + 7) - i + 1));
		strcpy(c->host_dir, c->url_scep + 7 + i);

                printf( "\thost: %s \n\tport: %s\n\tdir: %s\n",c->host,c->port,c->host_dir);
                return 1;
}

/* Aggiunge le estensioni in base al file di configurazione */
STACK_OF(X509_EXTENSION) * set_extensions(FILE * fp, STACK_OF(X509_EXTENSION) * exts) 
{
		 /* lettura del file config */
		char            buff[1024];
		int             i;
		exts = sk_X509_EXTENSION_new_null();
		while (fgets(buff, 1024, fp) && strncmp(buff, "[/X509v3EXT]", 12))
                      {
			 /* null-terminate: */
			buff[strlen(buff) - 1] = '\0';
			 /* skip leading white space: */
			for (i = 0; isspace(buff[i]); i++)
                 		;
   			 /* empty line? */
			if (!strlen(&buff[i]))
				continue;
			 /* comment? */
			if (!strncmp("#", &buff[i], 1))
				continue;
			for (; isspace(buff[i]); i++);
			if (!strncmp(buff, "key_usage", 9))
                        	{
   				add_ext(exts, NID_key_usage, get_string(&buff[i]));
		} else if (!strncmp(&buff[i], "subject_alt_name", 16))
		{
				add_ext(exts, NID_subject_alt_name, get_string(&buff[i]));
		} else if (!strncmp(&buff[i], "basic_constraints", 17))
		{
				add_ext(exts, NID_basic_constraints, get_string(&buff[i]));
		} else if (!strncmp(&buff[i], "netscape_cert_type", 18))
		{
				add_ext(exts, NID_netscape_cert_type, get_string(&buff[i]));
		} else if (!strncmp(&buff[i], "extended_key_usage", 18))
		{
				add_ext(exts, NID_ext_key_usage, get_string(&buff[i]));
		} else if (!strncmp(&buff[i], "netscape_comment", 16))
		{
				add_ext(exts, NID_netscape_comment, get_string(&buff[i]));
		}
	}
		return exts;

} 

int            
add_ext(STACK_OF(X509_REQUEST) * sk, int nid, char *value) 
{
	X509_EXTENSION * ex;
		ex = X509V3_EXT_conf_nid(NULL, NULL, nid, value);
		if (!ex)
			return 0;
		sk_X509_EXTENSION_push(sk, ex);
		return 1;
}


/*
 * Setup struttura del certificato dal file configurazione
 */ 
int            
set_cert_struct(FILE * fp, struct cert_data * c) 
{
	char            buff[1024];
	int             i;

 /* inizializzazione */
        c->cert_file = NULL;
 	c->key_file = NULL;
	c->CADN = NULL;
	c->err_flag = 0;
	c->req_command = NULL;
	c->challenge_password= NULL;
	c->renewed_flag = 99;
 /* lettura del file config */
		
        while (fgets(buff, 1024, fp) && strncmp(buff, "[/Certificate]", 14))
 	{
			 /* null-terminate: */
			buff[strlen(buff) - 1] = '\0';
			 /* skip leading white space: */
			for (i = 0; isspace(buff[i]); i++) ;
			 /* empty line? */
			if (!strlen(&buff[i]))
				continue;
			 /* comment? */
			if (!strncmp("#", &buff[i], 1))
				continue;
			if (!strncmp(&buff[i], "CADN", 4))
                 		{
				c->CADN = get_string(&buff[i]);
        		} else if (!strncmp(&buff[i], "Email", 5))
	        	{
				c->Email = get_string(&buff[i]);
		        } else if (!strncmp(&buff[i], "Country", 7))
		        {
				c->C = get_string(&buff[i]);
			} else if (!strncmp(&buff[i], "State", 5))
			{
				c->ST = get_string(&buff[i] + 5);
			} else if (!strncmp(&buff[i], "Location", 8))
			{
				c->L = get_string(&buff[i] + 8);
			} else if (!strncmp(&buff[i], "Organization", 12))
			{
                               c->O = get_string(&buff[i]);
                        } else if (!strncmp(&buff[i], "OrgUnit", 7))
			{
				c->OU = get_string(&buff[i]);
			} else if (!strncmp(&buff[i], "CommonName", 10))
		        {
				c->CN = get_string(&buff[i]);
			} else if (!strncmp(&buff[i], "CertFile", 8))
			{
				c->cert_file = get_string(&buff[i]);
			} else if (!strncmp(&buff[i], "KeyFile", 7))
			{
				c->key_file = get_string(&buff[i]);
			} else if (!strncmp(&buff[i], "[X509v3EXT]", 11))
			{
				c->exts = set_extensions(fp, c->exts);
			} else if (!strncmp(&buff[i], "ReqCommand", 10))
			{
				c->req_command = get_string (&buff[i]);
			}else if (!strncmp(&buff[i], "ChallengePassword", 17))
			{
				c->challenge_password = get_string (&buff[i]);
			}
                         else
			{
				fprintf(stderr, "%s: UNKNOWN OPTION: '%s'\n", pname, &buff[i]);
			}
			
	} 

        if (c->cert_file == NULL || c->key_file == NULL || c->CADN == NULL)
	{
			fprintf(stderr,
				"%s: BAD CONFIGURATION OF Certificate DATA (something is missing), check the config file\n", 
				pname);
			c->err_flag = ERR_CERT_BAD_CONFIG;
			return -1;
	}
        /* predispone il nome del file con la cartella per la lettura del Certificato  */

              	snprintf(temp_string,temp_string_len,"%s%s",cert_path,c->cert_file);
		c->cert = get_cert(temp_string);	/* legge il certificato dal file */
		return 1;

}

