
/*
 * Autosscep -- Simple SCEP client implementation
 * Copyright (c) Alberto Forino 2004. All rights reserved.
 * See the file COPYRIGHT for licensing information.
 */

#include "autosscep.h"
#include <math.h>
#include <string.h>

#define ENTRY_COUNT 7

/* struttura per le entry */
struct entry {
	char           *key;
	char           *value;
};


time_t
ASN1_UTCTIME_get(const ASN1_UTCTIME * s)
{
	struct tm       tm;
	int             offset;

	memset(&tm, '\0', sizeof tm);

#define g2(p) (((p)[0]-'0')*10+(p)[1]-'0')
	tm.tm_year = g2(s->data);
	if (tm.tm_year < 50)
		tm.tm_year += 100;
	tm.tm_mon = g2(s->data + 2) - 1;
	tm.tm_mday = g2(s->data + 4);
	tm.tm_hour = g2(s->data + 6);
	tm.tm_min = g2(s->data + 8);
	tm.tm_sec = g2(s->data + 10);
	if (s->data[12] == 'Z')
		offset = 0;
	else {
		offset = g2(s->data + 13) * 60 + g2(s->data + 15);
		if (s->data[12] == '-')
			offset = -offset;
	}
#undef g2

	return mktime(&tm) - offset * 60;	/* FIXME: mktime assumes the
						 * current timezone instead
						 * of UTC, and unless we
						 * rewrite OpenSSL in Lisp we
						 * cannot locally change the
						 * timezone without possibly
						 * interfering with other
						 * parts of the program.
						 * timegm, which uses UTC, is
						 * non-standard. Also time_t
						 * is inappropriate for
						 * general UTC times because
						 * it may a 32 bit type. */
}



/* Return 0 se il certificato non deve essere rinnovato */
int
check_date(X509 * cert, int ndays)
{
	int             days_for_expiration;
	double          sec;
	time_t          now, t_cert;
	struct tm      *t;
	ASN1_UTCTIME   *cert_time;

	if (cert == NULL)
		return 1;

	cert_time = X509_get_notAfter(cert);
	t_cert = ASN1_UTCTIME_get(cert_time);
	time(&now);
	t = gmtime(&now);
	sec = difftime(t_cert, now);
	days_for_expiration = sec / 86400;
	if (days_for_expiration > ndays) {
		return 0;
	} else {
		return 1;
	}
}

/*
 * Agisce sul vettore locale di strutture ca_data CAs cercandone una con il
 * DN che corrisponda a quello dell issuer specificato nel certificato
 * passato
 */
struct ca_data *
get_ca_struct(X509 * cert)
{
	int             i;
	char           *nome_issuer = NULL;

	if (cert == NULL)
		return NULL;

	/* prende il nome dell'issuer dal certificato */
	nome_issuer = X509_NAME_oneline(X509_get_issuer_name(cert), nome_issuer, 200);
	for (i = 0; i < ca_flag; i++) {

		if (!strcmp(CAs[i]->DN, nome_issuer)) {
			return CAs[i];
		}
	}

	return NULL;
}

/*
 * Genera una nuova chiave privata per una richiesta di certificato e la salva nella
 * directory delle chiavi con nome [nomefile].new
 *
 * usa le variabili globali:
 *	kbits ... numero di bit del modulo della nuova chiave
 *	key_path ... il percorso dove salvare il file
 */

int
gen_new_key(struct cert_data * c)
{
	RSA            *rsa_k = NULL;
	EVP_PKEY       *evp_k = EVP_PKEY_new();
	FILE           *fp;
	rsa_k = RSA_generate_key(kbits, //numero di bits
	     RSA_F4, //esponente(65537 come definito nelle librerie openssl)
				 NULL, //funzione di callback per la verifica
				 NULL);
	//argomenti funzione cb


		EVP_PKEY_assign_RSA(evp_k, rsa_k);

	/* assegna la chiave privata alla struttura */
		c->new_key = evp_k;

	/* scrive su file la nuova chiave privata */


	snprintf(temp_string,temp_string_len,"%s%s.new",key_path,c->key_file);

	
	if (!(fp = fopen(temp_string, "w"))) {
		//impostato da file conf
			fprintf(stderr, "%s: cannot open " "file for writing\n", pname);
		return -1;
	}
	PEM_write_PrivateKey(fp, evp_k, NULL, NULL, 0, NULL, NULL);
	fclose(fp);

	return 1;
}


/*
 * Genera una nuova richiesta di certificato e la salva nella
 * directory dei certificati con nome [nomefile].new
 *
 * usa le variabili globali:
 *
 *	cert_path ... il percorso dove salvare il file
 */

int
gen_new_req(struct cert_data * c)
{
	FILE           *fp;
	int             i;
	int             nid;
	X509_NAME_ENTRY *ent;
	X509_NAME      *subj = X509_NAME_new();
	struct entry    entries[ENTRY_COUNT] = {
		{"countryName", c->C},
		{"stateOrProvinceName", c->ST},
		{"localityName", c->L},
		{"organizationName", c->O},
		{"organizationalUnitName", c->OU},
		{"commonName", c->CN},
		{"emailAddress", c->Email}
	};
	c->req = X509_REQ_new();

	if (v_flag)
		fprintf(stdout, "%s: Creating request FROM CONFIG FILE DATA data for '%s'\n", pname, c->cert_file);
	if (c->cert == NULL && (c->C == NULL || c->ST == NULL || c->L == NULL || c->O == NULL || c->CN == NULL)) {
		fprintf(stderr, "%s: ERROR CREATING REQUEST FROM CONFIG FILE DATA, check them\n", pname);
		c->err_flag = ERR_CERT_BAD_CONFIG;
		return -1;
	}
	/* imposta il subject */



	for (i = 0; i < 7; i++) {
		if ((nid = OBJ_txt2nid(entries[i].key)) == NID_undef) {
			fprintf(stderr, "Error finding NID for %s\n", entries[i].key);
		}
		ent = X509_NAME_ENTRY_create_by_NID(NULL, nid, MBSTRING_ASC, entries[i].value, -1);

		X509_NAME_add_entry(subj, ent, -1, 0);
	}
	X509_REQ_set_subject_name(c->req, subj);


	/* imposta la versione */
	X509_REQ_set_version(c->req, 3L);

	/*
	 * Cambia la chiave pubblica dal cert mettendo quella corrispondente
	 * alla nuova k privata
	 */
	X509_REQ_set_pubkey(c->req, c->new_key);

	/* aggiunge le estensioni */

	X509_REQ_add_extensions(c->req, c->exts);

       /* aggiunge la challenge_password */
       if (c->challenge_password !=NULL) {
        X509_REQ_add1_attr_by_txt(c->req,"challengePassword", MBSTRING_UTF8 ,c->challenge_password, strlen(c->challenge_password));
       }

	/* Firma la req */
	X509_REQ_sign(c->req, c->new_key, sig_alg);

	/* scrive su file la nuova req */

        snprintf(temp_string,temp_string_len,"%s%s.csr",cert_path,c->cert_file);

	if (!(fp = fopen(temp_string, "w"))) {
		fprintf(stderr, "%s: cannot open " "file for writing\n", pname);
		return -1;
	}
	PEM_write_X509_REQ(fp, c->req);
	fclose(fp);

	if (d_flag)
		X509_REQ_print_fp(stdout, c->req);
  	return 1;
}

/*
 * Scrive il file di pending come [nomecert].pending
 *
 */
int
write_pending(struct cert_data * ce)
{
	FILE           *fp;
	snprintf(temp_string,temp_string_len,"%s%s.pending",cert_path,ce->cert_file);
 	if (!(fp = fopen(temp_string, "w"))) {
		fprintf(stderr, "%s: cannot open " "file for writing\n", pname);
		ce->err_flag = ERR_IO;
	}
	fprintf(fp, "%s\n", ce->transactionID);
	fclose(fp);
	return 1;
}

/*
 * Controlla se esiste il file di pending con il nome specificato e ritorna NULL se il file non c'
 * la stringa del transaction ID se il file c'
 */
char           *
ctrl_pending(struct cert_data *ce)
{
	FILE           *fp;
	char            temp[100];
	char           *ret;
        snprintf(temp_string,temp_string_len,"%s%s.pending",cert_path,ce->cert_file);
        fp = fopen(temp_string, "r");
	if (fp != NULL) {
		fgets(temp, 100, fp);
		fclose(fp);
		ret = malloc(sizeof(char) * strlen(temp));
		strcpy(ret, temp);
		ret[strlen(ret) - 1] = '\0';
		return ret;
	}
	return NULL;
}



/*
 * rinomina (se esiste) il file del certificato come [nomecertificato].OLD e
 * [nomecertificato].new in [nomecertificato]
 */
/* cos per la chiave */
int
update(struct cert_data * cert)
{
	char           *nome_cert, *nome_key, *new_name;
	FILE           *fp;

	/* IL CERTIFICATO */
	nome_cert = malloc(sizeof(char) * (strlen(cert_path) + strlen(cert->cert_file) + 1 + 4));
	strcpy(nome_cert, cert_path);
	strcat(nome_cert, cert->cert_file);

	/* rinomina cert in cert.OLD */
	new_name = malloc(sizeof(char) * (strlen(cert_path) + strlen(cert->cert_file) + 1 + 4));
	strcpy(new_name, nome_cert);
	strcat(new_name, ".OLD");
	/* controlla se c' gi il certificato */
	if ((fp = fopen(nome_cert, "r"))) {
		fclose(fp);
		rename(nome_cert, new_name);
	}
	free(new_name);
	/* rinomina cert.new in cert */
	new_name = malloc(sizeof(char) * (strlen(cert_path) + strlen(cert->cert_file) + 1 + 4));
	strcpy(new_name, nome_cert);
	strcat(new_name, ".new");
	rename(new_name, nome_cert);

	/* LA CHIAVE */
	nome_key = malloc(sizeof(char) * (strlen(key_path) + strlen(cert->key_file) + 1));
	strcpy(nome_key, key_path);
	strcat(nome_key, cert->key_file);

	/* rinomina cert in cert.OLD */
	new_name = malloc(sizeof(char) * (strlen(key_path) + strlen(cert->key_file) + 1 + 4));
	strcpy(new_name, nome_key);
	strcat(new_name, ".OLD");
	/* controlla se c' gi la chiave  */
	if ((fp = fopen(nome_key, "r"))) {
		fclose(fp);
		rename(nome_key, new_name);
		//eventualmente rinomina in key.OLD
	}
	free(new_name);
	/* rinomina key.new in key */
	new_name = malloc(sizeof(char) * (strlen(key_path) + strlen(cert->key_file) + 1 + 4));
	strcpy(new_name, nome_key);
	strcat(new_name, ".new");
	rename(new_name, nome_key);
	free(new_name);


	return 0;
}
