
/*
 * Autosscep -- Simple SCEP client implementation
 * Copyright (c) Alberto Forino 2004. All rights reserved.
 * See the file COPYRIGHT for licensing information.
 */


/* Misc. SCEP routines */

#include "autosscep.h"
#include "ias.h"

/*
 * Richiede alla CA il certificato della stessa e lo mette nella struttura
 */
void
request_ca_cert(struct ca_data * ca)
{
	unsigned int    n;
	char            http_string[16384];
	struct http_reply reply;
	int             c;
	unsigned char   md[EVP_MAX_MD_SIZE];
	BIO            *bp;
	if (v_flag)
		fprintf(stdout, "%s: CA file missing, trying request reom server\n", pname);


	/* Forge the HTTP message */
	if (ca->VHost_flag == 0)
		snprintf(http_string, sizeof(http_string),
			 "GET %s?operation=GetCACert&message=%s "
			 "HTTP/1.0\r\n\r\n", ca->host_dir, ca->ID == NULL ? "CAIdentifier" : ca->ID);
	else
		snprintf(http_string, sizeof(http_string),
			 "GET %s?operation=GetCACert&message=%s "
			 "HTTP/1.0\r\nHost: %s\r\n\r\n", ca->host_dir, ca->ID == NULL ? "CAIdentifier" : ca->ID, ca->host);

	printf("%s: requesting CA certificate\n", pname);
	if (d_flag)
		fprintf(stdout, "%s: scep msg: %s", pname, http_string);

	/*
	 * Send http message.
	 * Response is written to http_response struct "reply".
	 */

	reply.payload = NULL;
	if ((c = send_msg(&reply, http_string, ca->host,
			  atoi(ca->port), SCEP_OPERATION_GETCA)) == 1) {
		fprintf(stderr, "%s: error while sending "
			"message\n", pname);
		ca->err_flag = ERR_CA_FAILED_CACERT_REQ;
		return;
	}
	if (reply.payload == NULL) {
		fprintf(stderr, "%s: no data, perhaps you "
			"should define CA identifier (-i)\n", pname);
		ca->err_flag = ERR_CA_FAILED_CACERT_REQ;
		return;
	}
	printf("%s: valid response from server (reply: %d)\n", pname, reply.type);
	if (reply.type == SCEP_MIME_GETCA_RA) {
		/* XXXXXXXXXXXXXXXXXXXXX chain not verified */
		write_ca_ra(&reply, ca);
		return;
	}
	/* Read payload as DER X.509 object: */
	bp = BIO_new_mem_buf(reply.payload, reply.bytes);
	ca->cert = d2i_X509_bio(bp, NULL);
	ca->enc_cert = ca->cert;
	/* Read and print certificate information */
	if (!X509_digest(ca->cert, fp_alg, md, &n)) {
		ERR_print_errors_fp(stderr);
		return;
	}
	
	snprintf(temp_string, temp_string_len, "%s%s", ca_path, ca->cert_file);

	write_cert(ca->cert, temp_string);

}

/*
 * Initialize a SCEP transaction
 */
int
new_transaction(struct scep * s, struct cert_data * c)
{

	/* Set the whole struct as 0 */
	memset(s, 0, sizeof(*s));

	/* Set request and reply type */
	s->request_type = SCEP_REQUEST_NONE;
	s->request_type_str = NULL;
	s->reply_type = SCEP_REPLY_NONE;
	s->reply_type_str = NULL;
	s->pki_status = SCEP_PKISTATUS_UNSET;
	s->pki_status_str = NULL;
	s->fail_info_str = NULL;

	/* Set other variables */
	s->ias_getcertinit = pkcs7_issuer_and_subject_new();
	s->ias_getcert = PKCS7_ISSUER_AND_SERIAL_new();
	s->ias_getcrl = PKCS7_ISSUER_AND_SERIAL_new();

	/* Create transaction id */
	if (s->transaction_id == NULL)
		s->transaction_id = key_fingerprint(c->req);

	return (0);
}

/*
 * Crea un certificato autofirmato per firamre la transazione scep
 */
X509           *
new_selfsigned(struct cert_data * c)
{
	unsigned char  *ptr;
	X509           *cert;
	X509_NAME      *subject;
	ASN1_INTEGER   *serial;
	char           *temp;

	/* Get subject, issuer and extensions */
	if (!(subject = X509_REQ_get_subject_name(c->req))) {
		fprintf(stderr, "%s: error getting subject\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_BAD_REQ;
		return NULL;
	}
	/* Create new certificate */
	if (!(cert = X509_new())) {
		fprintf(stderr, "%s: error creating X509 certificate\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Set version (X509v3) */
	if (X509_set_version(cert, 2L) != 1) {
		fprintf(stderr, "%s: error setting cert version\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Get serial random no */

	temp = (char *) malloc(16);
	RAND_bytes(temp, 16);

	ptr = (unsigned char *) temp;
	if (!(serial = c2i_ASN1_INTEGER(NULL, &ptr, 32))) {
		fprintf(stderr, "%s: error converting serial\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	if (X509_set_serialNumber(cert, serial) != 1) {
		fprintf(stderr, "%s: error setting serial\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Set subject */
	if (X509_set_subject_name(cert, subject) != 1) {
		fprintf(stderr, "%s: error setting subject\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Set issuer (it's really the same as subject */
	if (X509_set_issuer_name(cert, subject) != 1) {
		fprintf(stderr, "%s: error setting issuer\n", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Set public key */
	if (X509_set_pubkey(cert, c->new_key) != 1) {
		fprintf(stderr, "%s: error setting public key", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Set duration */
	if (!(X509_gmtime_adj(X509_get_notBefore(cert), 0))) {
		fprintf(stderr, "%s: error setting begin time", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	if (!(X509_gmtime_adj(X509_get_notAfter(cert), SELFSIGNED_EXPIRE_DAYS * 24 * 60))) {
		fprintf(stderr, "%s: error setting end time", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	/* Sign certificate */
	if (!(X509_sign(cert, c->new_key, sig_alg))) {
		fprintf(stderr, "%s: error signing certificate", pname);
		ERR_print_errors_fp(stderr);
		c->err_flag = ERR_CERT_CREATING_SELFSIGNED;
		return NULL;
	}
	return cert;
}



/*
 * Initialize SCEP
 */
int
init_scep()
{
	unsigned char   randpool[1024];

	/* Add algorithms and init random pool */
	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();
	RAND_seed(randpool, sizeof(randpool));

	/* Create OpenSSL NIDs */

	nid_messageType = OBJ_create("2.16.840.1.113733.1.9.2", "messageType", "messageType");
	if (nid_messageType == 0) {
		goto err;
	}
	nid_pkiStatus = OBJ_create("2.16.840.1.113733.1.9.3", "pkiStatus", "pkiStatus");
	if (nid_pkiStatus == 0) {
		goto err;
	}
	nid_failInfo = OBJ_create("2.16.840.1.113733.1.9.4", "failInfo", "failInfo");
	if (nid_failInfo == 0) {
		goto err;
	}
	nid_senderNonce = OBJ_create("2.16.840.1.113733.1.9.5", "senderNonce", "senderNonce");
	if (nid_senderNonce == 0) {
		goto err;
	}
	nid_recipientNonce = OBJ_create("2.16.840.1.113733.1.9.6", "recipientNonce", "recipientNonce");
	if (nid_recipientNonce == 0) {
		goto err;
	}
	nid_transId = OBJ_create("2.16.840.1.113733.1.9.7", "transId", "transId");
	if (nid_transId == 0) {
		goto err;
	}
	nid_extensionReq = OBJ_create("2.16.840.1.113733.1.9.8", "extensionReq", "extensionReq");
	if (nid_extensionReq == 0) {
		goto err;
	}
	return (0);

err:
	fprintf(stderr, "%s: cannot create OID\n", pname);
	return (1);

}

/*
 * Calculate transaction id.
 * Return pointer to ascii presentation of the hash.
 */
char           *
key_fingerprint(X509_REQ * req)
{
	char           *ret, *str;
	unsigned char  *data, md[MD5_DIGEST_LENGTH];
	int             c, len;
	BIO            *bio;
	MD5_CTX         ctx;

	/* Assign space for ASCII presentation of the digest */
	str = (unsigned char *) malloc(2 * MD5_DIGEST_LENGTH + 1);
	ret = str;

	/* Create new memory bio for reading the public key */
	bio = BIO_new(BIO_s_mem());
	i2d_PUBKEY_bio(bio, X509_REQ_get_pubkey(req));
	len = BIO_get_mem_data(bio, &data);

	/* Calculate MD5 hash: */
	MD5_Init(&ctx);
	MD5_Update(&ctx, data, len);
	MD5_Final(md, &ctx);

	/* Copy as ASCII string and return: */
	for (c = 0; c < MD5_DIGEST_LENGTH; c++, str += 2) {
		sprintf((char *) str, "%02X", md[c]);

	}
	*(str + 2) = '\0';

	return (ret);
}
